@@ -11,6 +11,7 @@ GET_APPS = "com.webos.applicationManager/listLaunchPoints"
 GET_APP_STATUS = "com.webos.service.appstatus/getAppStatus"
 SEND_ENTER = "com.webos.service.ime/sendEnterKey"
 SEND_DELETE = "com.webos.service.ime/deleteCharacters"
+INSERT_TEXT = "com.webos.service.ime/insertText"
 SET_3D_ON = "com.webos.service.tv.display/set3DOn"
 SET_3D_OFF = "com.webos.service.tv.display/set3DOff"
 GET_SOFTWARE_INFO = "com.webos.service.update/getCurrentSWInformation"
@@ -43,3 +44,5 @@ CALIBRATION = "externalpq/setExternalPqData"
 GET_SOUND_OUTPUT = "com.webos.service.apiadapter/audio/getSoundOutput"
 CHANGE_SOUND_OUTPUT = "com.webos.service.apiadapter/audio/changeSoundOutput"
 GET_POWER_STATE = "com.webos.service.tvpower/power/getPowerState"
+TURN_OFF_SCREEN = "com.webos.service.tvpower/power/turnOffScreen"
+TURN_ON_SCREEN = "com.webos.service.tvpower/power/turnOnScreen"
