@@ -1,24 +1,32 @@
 import asyncio
 import base64
 import copy
+import functools
 import json
 import logging
 import os
 
 import numpy as np
 import websockets
+from sqlitedict import SqliteDict
 
 from . import buttons as btn
 from . import cal_commands as cal
 from . import endpoints as ep
-from .constants import CALIBRATION_TYPE_MAP, DEFAULT_CAL_DATA
+from .constants import BT2020_PRIMARIES, CALIBRATION_TYPE_MAP, DEFAULT_CAL_DATA
 from .handshake import REGISTRATION_MESSAGE
-from .lut_tools import read_cal_file, read_cube_file, unity_lut_1d, unity_lut_3d
+from .lut_tools import (
+    create_dolby_vision_config,
+    read_cal_file,
+    read_cube_file,
+    unity_lut_1d,
+    unity_lut_3d,
+)
 
 logger = logging.getLogger(__name__)
 
 
-KEY_FILE_NAME = ".aiopylgtv"
+KEY_FILE_NAME = ".aiopylgtv.sqlite"
 USER_HOME = "HOME"
 
 
@@ -32,17 +40,36 @@ class PyLGTVCmdException(Exception):
         self.message = message
 
 
+class PyLGTVCmdError(PyLGTVCmdException):
+    def __init__(self, message):
+        self.message = message
+
+
+class PyLGTVServiceNotFoundError(PyLGTVCmdError):
+    def __init__(self, message):
+        self.message = message
+
+
 class WebOsClient:
-    def __init__(self, ip, key_file_path=None, timeout_connect=2, ping_interval=1):
+    def __init__(
+        self,
+        ip,
+        key_file_path=None,
+        timeout_connect=2,
+        ping_interval=1,
+        ping_timeout=20,
+        client_key=None,
+    ):
         """Initialize the client."""
         self.ip = ip
         self.port = 3000
         self.key_file_path = key_file_path
-        self.client_key = None
+        self.client_key = client_key
         self.web_socket = None
         self.command_count = 0
         self.timeout_connect = timeout_connect
         self.ping_interval = ping_interval
+        self.ping_timeout = ping_timeout
         self.connect_task = None
         self.connect_result = None
         self.connection = None
@@ -64,7 +91,18 @@ class WebOsClient:
         self.state_update_callbacks = []
         self.doStateUpdate = False
 
-        self.load_key_file()
+    @classmethod
+    async def create(cls, *args, **kwargs):
+        client = cls(*args, **kwargs)
+        await client.async_init()
+        return client
+
+    async def async_init(self):
+        """Load client key from config file if in use."""
+        if self.client_key is None:
+            self.client_key = await asyncio.get_running_loop().run_in_executor(
+                None, self.read_client_key
+            )
 
     @staticmethod
     def _get_key_file_path():
@@ -76,28 +114,20 @@ class WebOsClient:
 
         return os.path.join(os.getcwd(), KEY_FILE_NAME)
 
-    def load_key_file(self):
+    def read_client_key(self):
         """Try to load the client key for the current ip."""
-        self.client_key = None
+
         if self.key_file_path:
             key_file_path = self.key_file_path
         else:
             key_file_path = self._get_key_file_path()
-        key_dict = {}
 
         logger.debug("load keyfile from %s", key_file_path)
 
-        if os.path.isfile(key_file_path):
-            with open(key_file_path, "r") as f:
-                raw_data = f.read()
-                if raw_data:
-                    key_dict = json.loads(raw_data)
+        with SqliteDict(key_file_path) as conf:
+            return conf.get(self.ip)
 
-        logger.debug("getting client_key for %s from %s", self.ip, key_file_path)
-        if self.ip in key_dict:
-            self.client_key = key_dict[self.ip]
-
-    def save_key_file(self):
+    def write_client_key(self):
         """Save the current client key."""
         if self.client_key is None:
             return
@@ -109,24 +139,9 @@ class WebOsClient:
 
         logger.debug("save keyfile to %s", key_file_path)
 
-        if os.path.isfile(key_file_path):
-            with open(key_file_path, "r+") as f:
-                raw_data = f.read()
-                f.seek(0)
-                f.truncate()
-                key_dict = {}
-
-                if raw_data:
-                    key_dict = json.loads(raw_data)
-
-                key_dict[self.ip] = self.client_key
-
-                f.write(json.dumps(key_dict))
-        else:
-            with open(key_file_path, "w") as f:
-                key_dict = {}
-                key_dict[self.ip] = self.client_key
-                f.write(json.dumps(key_dict))
+        with SqliteDict(key_file_path) as conf:
+            conf[self.ip] = self.client_key
+            conf.commit()
 
     async def connect(self):
         if not self.is_connected():
@@ -167,6 +182,7 @@ class WebOsClient:
                     f"ws://{self.ip}:{self.port}",
                     ping_interval=None,
                     close_timeout=self.timeout_connect,
+                    max_size=None,
                 ),
                 timeout=self.timeout_connect,
             )
@@ -182,7 +198,9 @@ class WebOsClient:
                 response = json.loads(raw_response)
                 if response["type"] == "registered":
                     self.client_key = response["payload"]["client-key"]
-                    self.save_key_file()
+                    await asyncio.get_running_loop().run_in_executor(
+                        None, self.write_client_key
+                    )
 
             if not self.client_key:
                 raise PyLGTVPairException("Unable to pair")
@@ -197,7 +215,9 @@ class WebOsClient:
             )
             if self.ping_interval is not None:
                 handler_tasks.add(
-                    asyncio.create_task(self.ping_handler(ws, self.ping_interval))
+                    asyncio.create_task(
+                        self.ping_handler(ws, self.ping_interval, self.ping_timeout)
+                    )
                 )
             self.connection = ws
 
@@ -218,7 +238,11 @@ class WebOsClient:
             handler_tasks.add(asyncio.create_task(inputws.wait_closed()))
             if self.ping_interval is not None:
                 handler_tasks.add(
-                    asyncio.create_task(self.ping_handler(inputws, self.ping_interval))
+                    asyncio.create_task(
+                        self.ping_handler(
+                            inputws, self.ping_interval, self.ping_timeout
+                        )
+                    )
                 )
             self.input_connection = inputws
 
@@ -241,11 +265,11 @@ class WebOsClient:
             subscribe_tasks = set()
             for coro in subscribe_coros:
                 subscribe_tasks.add(asyncio.create_task(coro))
-            subscribe_done, _ = await asyncio.wait(subscribe_tasks)
-            for task in subscribe_done:
+            await asyncio.wait(subscribe_tasks)
+            for task in subscribe_tasks:
                 try:
                     task.result()
-                except PyLGTVCmdException:
+                except PyLGTVServiceNotFoundError:
                     pass
             # set placeholder power state if not available
             if not self._power_state:
@@ -307,14 +331,15 @@ class WebOsClient:
                     except asyncio.CancelledError:
                         pass
 
-    async def ping_handler(self, ws, interval):
+    async def ping_handler(self, ws, interval, timeout):
         try:
             while True:
                 await asyncio.sleep(interval)
                 # In the "Suspend" state the tv can keep a connection alive, but will not respond to pings
                 if self._power_state.get("state") != "Suspend":
                     ping_waiter = await ws.ping()
-                    await asyncio.wait_for(ping_waiter, timeout=self.timeout_connect)
+                    if timeout is not None:
+                        await asyncio.wait_for(ping_waiter, timeout=timeout)
         except (
             asyncio.TimeoutError,
             asyncio.CancelledError,
@@ -369,11 +394,11 @@ class WebOsClient:
             if tasks:
                 closeout_task = asyncio.create_task(asyncio.wait(tasks))
 
-            while not closeout_task.done():
-                try:
-                    await asyncio.shield(closeout_task)
-                except asyncio.CancelledError:
-                    pass
+                while not closeout_task.done():
+                    try:
+                        await asyncio.shield(closeout_task)
+                    except asyncio.CancelledError:
+                        pass
 
     # manage state
     @property
@@ -438,6 +463,12 @@ class WebOsClient:
         else:
             return True
 
+    @property
+    def is_screen_on(self):
+        if self.is_on:
+            return self._power_state.get("state") != "Screen Off"
+        return False
+
     def calibration_support_info(self):
         info = {
             "lut1d": False,
@@ -569,10 +600,19 @@ class WebOsClient:
         if self.state_update_callbacks and self.doStateUpdate:
             await self.do_state_update_callbacks()
 
-    async def set_apps_state(self, apps):
-        self._apps = {}
-        for app in apps:
-            self._apps[app["id"]] = app
+    async def set_apps_state(self, payload):
+        apps = payload.get("launchPoints")
+        if apps is not None:
+            self._apps = {}
+            for app in apps:
+                self._apps[app["id"]] = app
+        else:
+            change = payload["change"]
+            app_id = payload["id"]
+            if change == "removed":
+                del self._apps[app_id]
+            else:
+                self._apps[app_id] = payload
 
         if self.state_update_callbacks and self.doStateUpdate:
             await self.do_state_update_callbacks()
@@ -640,7 +680,13 @@ class WebOsClient:
 
         returnValue = payload.get("returnValue") or payload.get("subscribed")
 
-        if returnValue is None:
+        if response.get("type") == "error":
+            error = response.get("error")
+            if error == "404 no such service or method":
+                raise PyLGTVServiceNotFoundError(error)
+            else:
+                raise PyLGTVCmdError(error)
+        elif returnValue is None:
             raise PyLGTVCmdException(f"Invalid request response {response}")
         elif not returnValue:
             raise PyLGTVCmdException(f"Request failed with response {response}")
@@ -727,11 +773,7 @@ class WebOsClient:
 
     async def subscribe_apps(self, callback):
         """Subscribe to changes in available apps."""
-
-        async def apps(payload):
-            await callback(payload.get("launchPoints"))
-
-        return await self.subscribe(apps, ep.GET_APPS)
+        return await self.subscribe(callback, ep.GET_APPS)
 
     async def get_current_app(self):
         """Get the current app id."""
@@ -791,6 +833,14 @@ class WebOsClient:
         """Play media."""
         return await self.request(ep.POWER_ON)
 
+    async def turn_screen_off(self):
+        """Turn TV Screen off."""
+        await self.command("request", ep.TURN_OFF_SCREEN)
+
+    async def turn_screen_on(self):
+        """Turn TV Screen on."""
+        await self.command("request", ep.TURN_ON_SCREEN)
+
     # 3D Mode
     async def turn_3d_on(self):
         """Turn 3D on."""
@@ -847,13 +897,13 @@ class WebOsClient:
     async def get_volume(self):
         """Get the current volume."""
         res = await self.request(ep.GET_VOLUME)
-        return res.get("volume")
+        return res.get("volumeStatus", res).get("volume")
 
     async def subscribe_volume(self, callback):
         """Subscribe to changes in the current volume."""
 
         async def volume(payload):
-            await callback(payload.get("volume"))
+            await callback(payload.get("volumeStatus", payload).get("volume"))
 
         return await self.subscribe(volume, ep.GET_VOLUME)
 
@@ -963,6 +1013,11 @@ class WebOsClient:
         """Send delete key."""
         return await self.request(ep.SEND_DELETE)
 
+    # Text entry
+    async def insert_text(self, text, replace=False):
+        """Insert text into field, optionally replace existing text."""
+        return await self.request(ep.INSERT_TEXT, {"text": text, "replace": replace})
+
     # Web
     async def open_url(self, url):
         """Open URL."""
@@ -1070,11 +1125,15 @@ class WebOsClient:
 
     def validateCalibrationData(self, data, shape, dtype):
         if not isinstance(data, np.ndarray):
-            raise TypeError
+            raise TypeError(f"data must be of type ndarray but is instead {type(data)}")
         if data.shape != shape:
-            raise ValueError
+            raise ValueError(
+                f"data should have shape {shape} but instead has {data.shape}"
+            )
         if data.dtype != dtype:
-            raise TypeError
+            raise TypeError(
+                f"numpy dtype should be {dtype} but is instead {data.dtype}"
+            )
 
     async def calibration_request(self, command, picMode, data):
         dataenc = base64.b64encode(data.tobytes()).decode()
@@ -1087,8 +1146,9 @@ class WebOsClient:
             "dataType": CALIBRATION_TYPE_MAP[data.dtype.name],
             "profileNo": 0,
             "programID": 1,
-            "picMode": picMode,
         }
+        if picMode is not None:
+            payload["picMode"] = picMode
 
         return await self.request(ep.CALIBRATION, payload)
 
@@ -1108,7 +1168,7 @@ class WebOsClient:
                 f"1D LUT Upload not supported by tv model {model}."
             )
         if data is None:
-            data = unity_lut_1d()
+            data = await asyncio.get_running_loop().run_in_executor(None, unity_lut_1d)
         self.validateCalibrationData(data, (3, 1024), np.uint16)
         return await self.calibration_request(cal.UPLOAD_1D_LUT, picMode, data)
 
@@ -1123,7 +1183,9 @@ class WebOsClient:
                 f"3D LUT Upload not supported by tv model {model}."
             )
         if data is None:
-            data = unity_lut_3d(lut3d_size)
+            data = await asyncio.get_running_loop().run_in_executor(
+                None, unity_lut_3d, lut3d_size
+            )
         lut3d_shape = (lut3d_size, lut3d_size, lut3d_size, 3)
         self.validateCalibrationData(data, lut3d_shape, np.uint16)
         return await self.calibration_request(command, picMode, data)
@@ -1135,9 +1197,6 @@ class WebOsClient:
         return await self.upload_3d_lut(cal.UPLOAD_3D_LUT_BT2020, picMode, data)
 
     async def set_ui_data(self, command, picMode, value):
-        if isinstance(value, str):
-            value = int(value)
-
         if not (value >= 0 and value <= 100):
             raise ValueError
 
@@ -1180,6 +1239,35 @@ class WebOsClient:
         self.validateCalibrationData(data, (3, 3), np.float32)
         return await self.calibration_request(cal.BT2020_3BY3_GAMUT_DATA, picMode, data)
 
+    async def set_dolby_vision_config_data(
+        self, white_level=700.0, black_level=0.0, gamma=2.2, primaries=BT2020_PRIMARIES
+    ):
+
+        info = self.calibration_support_info()
+        dv_config_type = info["dv_config_type"]
+        if dv_config_type is None:
+            model = self._system_info["modelName"]
+            raise PyLGTVCmdException(
+                f"Dolby Vision Configuration Upload not supported by tv model {model}."
+            )
+
+        config = await asyncio.get_running_loop().run_in_executor(
+            None,
+            functools.partial(
+                create_dolby_vision_config,
+                version=dv_config_type,
+                white_level=white_level,
+                black_level=black_level,
+                gamma=gamma,
+                primaries=primaries,
+            ),
+        )
+
+        data = np.frombuffer(config.encode(), dtype=np.uint8)
+        return await self.calibration_request(
+            command=cal.DOLBY_CFG_DATA, picMode=None, data=data
+        )
+
     async def set_tonemap_params(
         self,
         picMode,
@@ -1208,18 +1296,10 @@ class WebOsClient:
         return await self.calibration_request(cal.SET_TONEMAP_PARAM, picMode, data)
 
     async def ddc_reset(self, picMode, reset_1d_lut=True):
-        if isinstance(reset_1d_lut, str):
-            if reset_1d_lut.lower() == "true":
-                reset_1d_lut = True
-            elif reset_1d_lut.lower() == "false":
-                reset_1d_lut = False
-            else:
-                try:
-                    reset_1d_lut = bool(int(reset_1d_lut))
-                except ValueError:
-                    raise ValueError(
-                        f"Invalid parameter {reset_1d_lut} for ddc_reset, should be a boolean."
-                    )
+        if not isinstance(reset_1d_lut, bool):
+            raise TypeError(
+                f"reset_1d_lut should be a bool, instead got {reset_1d_lut} of type {type(reset_1d_lut)}."
+            )
 
         await self.set_1d_2_2_en(picMode)
         await self.set_1d_0_45_en(picMode)
@@ -1242,9 +1322,13 @@ class WebOsClient:
     async def upload_1d_lut_from_file(self, picMode, filename):
         ext = filename.split(".")[-1].lower()
         if ext == "cal":
-            lut = read_cal_file(filename)
+            lut = await asyncio.get_running_loop().run_in_executor(
+                None, read_cal_file, filename
+            )
         elif ext == "cube":
-            lut = read_cube_file(filename)
+            lut = await asyncio.get_running_loop().run_in_executor(
+                None, read_cube_file, filename
+            )
         else:
             raise ValueError(
                 f"Unsupported file format {ext} for 1D LUT.  Supported file formats are cal and cube."
@@ -1255,7 +1339,9 @@ class WebOsClient:
     async def upload_3d_lut_from_file(self, command, picMode, filename):
         ext = filename.split(".")[-1].lower()
         if ext == "cube":
-            lut = read_cube_file(filename)
+            lut = await asyncio.get_running_loop().run_in_executor(
+                None, read_cube_file, filename
+            )
         else:
             raise ValueError(
                 f"Unsupported file format {ext} for 3D LUT.  Supported file formats are cube."
